"""This file was generated by xsdata, v24.1, on 2024-04-29 12:10:47

Generator: DataclassGenerator
See: https://xsdata.readthedocs.io/
"""
from eflips.ingest.vdv452data.vdv452_v1_5 import (
    BasisVerGueltigkeit as BasisVerGueltigkeit,
    Firmenkalender as Firmenkalender,
    LidVerlauf as LidVerlauf,
    MengeFzgTyp as MengeFzgTyp,
    OrtHztf as OrtHztf,
    RecFrt as RecFrt,
    RecFrtHzt as RecFrtHzt,
    RecLid as RecLid,
    RecOrt as RecOrt,
    RecSel as RecSel,
    RecUeb as RecUeb,
    RecUmlauf as RecUmlauf,
    SelFztFeld as SelFztFeld,
    UebFzt as UebFzt,
    VdvBaseObject as VdvBaseObject,
)
