"""This file was generated by xsdata, v24.1, on 2024-01-25 17:07:33

Generator: DataclassGenerator
See: https://xsdata.readthedocs.io/
"""
from eflips.ingest.legacy.xmldata.bvg_xml import (
    ErgebnisReturnCode,
    FahrtFahrgastrelevant,
    Linienfahrplan,
    MeldungMeldungskategorie,
    MeldungMeldungsnummer,
    MeldungMeldungstext,
    NetzpunktNetzpunkttyp,
    NetzpunktMitFahrgastwechsel,
    ParameterName,
    PunktFahrgastwechsel,
    PunktVeroeffentlicht,
    RouteHauptroute,
    RouteRichtung,
    T41Veroeffentlicht,
    ZwischenpunktEntfernungVomStart,
    ZwischenpunktNetzpunkt,
)

__all__ = [
    "ErgebnisReturnCode",
    "FahrtFahrgastrelevant",
    "Linienfahrplan",
    "MeldungMeldungskategorie",
    "MeldungMeldungsnummer",
    "MeldungMeldungstext",
    "NetzpunktNetzpunkttyp",
    "NetzpunktMitFahrgastwechsel",
    "ParameterName",
    "PunktFahrgastwechsel",
    "PunktVeroeffentlicht",
    "RouteHauptroute",
    "RouteRichtung",
    "T41Veroeffentlicht",
    "ZwischenpunktEntfernungVomStart",
    "ZwischenpunktNetzpunkt",
]
